from django.conf import settings
from django.db import models


class OrderItemManager(models.Manager):
    def create_from_cart_item(self, item, *, order: settings.SHOP_ORDER_MODEL):
        return self.create(
            order=order,
            product=item.product,
            quantity=item.quantity,
            unit_price_paid=item.product.price,
        )


class BaseOrderItem(models.Model):
    class Meta:
        abstract = True

    objects = OrderItemManager()

    order = models.ForeignKey(
        settings.SHOP_ORDER_MODEL, related_name="items", on_delete=models.PROTECT
    )
    product = models.ForeignKey(settings.SHOP_PRODUCT_MODEL, on_delete=models.PROTECT)

    unit_price_paid = models.DecimalField(max_digits=7, decimal_places=2)
    quantity = models.PositiveIntegerField(default=0)

    @property
    def total_price(self):
        return self.unit_price_paid * self.quantity

    def clean(self, *args, **kwargs):
        super().clean()

        if self.quantity <= 0:
            raise ValidationError("Quantity of OrderItem must be greater than zero.")

    def save(self, *args, **kwargs):
        self.full_clean()
        super().save()



from django.conf import settings
from django.core.exceptions import ValidationError
from django.db import models


class BaseCartItem(models.Model):
    class Meta:
        abstract = True

    cart = models.ForeignKey(
        settings.SHOP_CART_MODEL, related_name="items", on_delete=models.CASCADE
    )

    quantity = models.PositiveIntegerField(default=0)
    product = models.ForeignKey(settings.SHOP_PRODUCT_MODEL, on_delete=models.PROTECT)

    def __str__(self):
        return f"CartItem ({self.product.name} x {self.quantity})"

    @property
    def total_price(self):
        return self.product.price * self.quantity  # pylint: disable=no-member

    def clean(self, *args, **kwargs):
        super().clean()

        if self.quantity <= 0:
            raise ValidationError("Quantity of CartItem must be greater than zero.")

    def save(self, *args, **kwargs):
        self.full_clean()
        super().save()



from typing import Optional
from django.db import models

from ..concrete.address import Address
from ..concrete.shipping_option import ShippingOption


class BaseCheckoutDetails(models.Model):
    class Meta:
        abstract = True

    customer_name = models.CharField(max_length=100)
    customer_company = models.CharField(max_length=100, blank=True)
    customer_email = models.EmailField()
    customer_phone = models.CharField(max_length=100)
    customer_coupon = models.CharField(max_length=100, blank=True)

    shipping_selection = models.ForeignKey(
        ShippingOption, null=True, on_delete=models.PROTECT
    )

    shipping_address = models.ForeignKey(
        Address, related_name="+", on_delete=models.PROTECT
    )

    billing_address_same_as_shipping = models.BooleanField(default=True)

    billing_name = models.CharField(max_length=100, blank=True)
    billing_company = models.CharField(max_length=100, blank=True)
    billing_address = models.ForeignKey(
        Address, null=True, related_name="+", on_delete=models.PROTECT
    )

    def __str__(self):
        if self.customer_name:
            return f"CheckoutDetails ({self.customer_name})"

        return super().__str__()



from django.db import migrations


def forwards_func(apps, schema_editor):
    Country = apps.get_model("django_base_shop", "Country")

    db_alias = schema_editor.connection.alias
    Country.objects.using(db_alias).bulk_create(
        [
            Country(name="Afghanistan"),
            Country(name="Albania"),
            Country(name="Algeria"),
            Country(name="Andorra"),
            Country(name="Angola"),
            Country(name="Antigua & Deps"),
            Country(name="Argentina"),
            Country(name="Armenia"),
            Country(name="Australia"),
            Country(name="Austria"),
            Country(name="Azerbaijan"),
            Country(name="Bahamas"),
            Country(name="Bahrain"),
            Country(name="Bangladesh"),
            Country(name="Barbados"),
            Country(name="Belarus"),
            Country(name="Belgium"),
            Country(name="Belize"),
            Country(name="Benin"),
            Country(name="Bhutan"),
            Country(name="Bolivia"),
            Country(name="Bosnia Herzegovina"),
            Country(name="Botswana"),
            Country(name="Brazil"),
            Country(name="Brunei"),
            Country(name="Bulgaria"),
            Country(name="Burkina"),
            Country(name="Burundi"),
            Country(name="Cambodia"),
            Country(name="Cameroon"),
            Country(name="Canada"),
            Country(name="Cape Verde"),
            Country(name="Central African Rep"),
            Country(name="Chad"),
            Country(name="Chile"),
            Country(name="China"),
            Country(name="Colombia"),
            Country(name="Comoros"),
            Country(name="Congo"),
            Country(name="Democratic Republic of Congo"),
            Country(name="Costa Rica"),
            Country(name="Croatia"),
            Country(name="Cuba"),
            Country(name="Cyprus"),
            Country(name="Czech Republic"),
            Country(name="Denmark"),
            Country(name="Djibouti"),
            Country(name="Dominica"),
            Country(name="Dominican Republic"),
            Country(name="East Timor"),
            Country(name="Ecuador"),
            Country(name="Egypt"),
            Country(name="El Salvador"),
            Country(name="Equatorial Guinea"),
            Country(name="Eritrea"),
            Country(name="Estonia"),
            Country(name="Ethiopia"),
            Country(name="Fiji"),
            Country(name="Finland"),
            Country(name="France"),
            Country(name="Gabon"),
            Country(name="Gambia"),
            Country(name="Georgia"),
            Country(name="Germany"),
            Country(name="Ghana"),
            Country(name="Greece"),
            Country(name="Grenada"),
            Country(name="Guatemala"),
            Country(name="Guinea"),
            Country(name="Guinea-Bissau"),
            Country(name="Guyana"),
            Country(name="Haiti"),
            Country(name="Honduras"),
            Country(name="Hungary"),
            Country(name="Iceland"),
            Country(name="India"),
            Country(name="Indonesia"),
            Country(name="Iran"),
            Country(name="Iraq"),
            Country(name="Ireland "),
            Country(name="Israel"),
            Country(name="Italy"),
            Country(name="Ivory Coast"),
            Country(name="Jamaica"),
            Country(name="Japan"),
            Country(name="Jordan"),
            Country(name="Kazakhstan"),
            Country(name="Kenya"),
            Country(name="Kiribati"),
            Country(name="Korea North"),
            Country(name="Korea South"),
            Country(name="Kosovo"),
            Country(name="Kuwait"),
            Country(name="Kyrgyzstan"),
            Country(name="Laos"),
            Country(name="Latvia"),
            Country(name="Lebanon"),
            Country(name="Lesotho"),
            Country(name="Liberia"),
            Country(name="Libya"),
            Country(name="Liechtenstein"),
            Country(name="Lithuania"),
            Country(name="Luxembourg"),
            Country(name="Macedonia"),
            Country(name="Madagascar"),
            Country(name="Malawi"),
            Country(name="Malaysia"),
            Country(name="Maldives"),
            Country(name="Mali"),
            Country(name="Malta"),
            Country(name="Marshall Islands"),
            Country(name="Mauritania"),
            Country(name="Mauritius"),
            Country(name="Mexico"),
            Country(name="Micronesia"),
            Country(name="Moldova"),
            Country(name="Monaco"),
            Country(name="Mongolia"),
            Country(name="Montenegro"),
            Country(name="Morocco"),
            Country(name="Mozambique"),
            Country(name="Myanmar"),
            Country(name="Namibia"),
            Country(name="Nauru"),
            Country(name="Nepal"),
            Country(name="Netherlands"),
            Country(name="New Zealand"),
            Country(name="Nicaragua"),
            Country(name="Niger"),
            Country(name="Nigeria"),
            Country(name="Norway"),
            Country(name="Oman"),
            Country(name="Pakistan"),
            Country(name="Palau"),
            Country(name="Panama"),
            Country(name="Papua New Guinea"),
            Country(name="Paraguay"),
            Country(name="Peru"),
            Country(name="Philippines"),
            Country(name="Poland"),
            Country(name="Portugal"),
            Country(name="Qatar"),
            Country(name="Romania"),
            Country(name="Russian Federation"),
            Country(name="Rwanda"),
            Country(name="St Kitts & Nevis"),
            Country(name="St Lucia"),
            Country(name="Saint Vincent & the Grenadines"),
            Country(name="Samoa"),
            Country(name="San Marino"),
            Country(name="Sao Tome & Principe"),
            Country(name="Saudi Arabia"),
            Country(name="Senegal"),
            Country(name="Serbia"),
            Country(name="Seychelles"),
            Country(name="Sierra Leone"),
            Country(name="Singapore"),
            Country(name="Slovakia"),
            Country(name="Slovenia"),
            Country(name="Solomon Islands"),
            Country(name="Somalia"),
            Country(name="South Africa"),
            Country(name="South Sudan"),
            Country(name="Spain"),
            Country(name="Sri Lanka"),
            Country(name="Sudan"),
            Country(name="Suriname"),
            Country(name="Swaziland"),
            Country(name="Sweden"),
            Country(name="Switzerland"),
            Country(name="Syria"),
            Country(name="Taiwan"),
            Country(name="Tajikistan"),
            Country(name="Tanzania"),
            Country(name="Thailand"),
            Country(name="Togo"),
            Country(name="Tonga"),
            Country(name="Trinidad & Tobago"),
            Country(name="Tunisia"),
            Country(name="Turkey"),
            Country(name="Turkmenistan"),
            Country(name="Tuvalu"),
            Country(name="Uganda"),
            Country(name="Ukraine"),
            Country(name="United Arab Emirates"),
            Country(name="United Kingdom"),
            Country(name="United States"),
            Country(name="Uruguay"),
            Country(name="Uzbekistan"),
            Country(name="Vanuatu"),
            Country(name="Vatican City"),
            Country(name="Venezuela"),
            Country(name="Vietnam"),
            Country(name="Yemen"),
            Country(name="Zambia"),
            Country(name="Zimbabwe"),
        ]
    )


def reverse_func(apps, schema_editor):
    Country = apps.get_model("django_base_shop", "Country")

    db_alias = schema_editor.connection.alias
    Country.objects.all().delete()


class Migration(migrations.Migration):

    dependencies = [
        ("django_base_shop", "0001_initial"),
    ]

    operations = [
        migrations.RunPython(forwards_func, reverse_func),
    ]



from django.db import models


class Image(models.Model):
    handle = models.CharField(max_length=200, unique=True, db_index=True)
    image = models.ImageField(upload_to="images/")

    def __str__(self):
        return self.handle



from decimal import Decimal
from functools import partial
from secrets import token_hex

from django.db import models
from django.conf import settings
from django.apps import apps

from ..abstract.base_cart import BaseCart


class OrderManager(models.Manager):
    def create_from_cart(self, cart: BaseCart):

        if cart.checkout_details.shipping_selection is None:
            raise ValueError(
                "Cart must contain a shipping option when creating an Order."
            )

        order = self.create(
            checkout_details=cart.checkout_details,
            shipping_paid=cart.checkout_details.shipping_selection.price,
            subtotal_paid=cart.subtotal,
        )

        OrderItem = apps.get_model(settings.SHOP_ORDER_ITEM_MODEL)

        for item in cart.items.all():
            OrderItem.objects.create_from_cart_item(item, order=order)

        return order


class BaseOrder(models.Model):
    class Meta:
        abstract = True
        indexes = (models.Index(fields=["order_token"]),)

    objects = OrderManager()

    # Meta info
    date_placed = models.DateTimeField(auto_now_add=True)
    order_token = models.CharField(
        max_length=32, unique=True, editable=False, default=partial(token_hex, 16)
    )

    checkout_details = models.ForeignKey(
        settings.SHOP_CHECKOUT_DETAILS_MODEL, related_name="+", on_delete=models.PROTECT
    )

    # Snapshot price information
    #   This is done incase price of FK field is updated.
    subtotal_paid = models.DecimalField(max_digits=7, decimal_places=2)
    shipping_paid = models.DecimalField(max_digits=7, decimal_places=2)

    def __len__(self):
        return self.items.count()

    def __str__(self):
        return f"Order ({self.checkout_details.customer_name})"

    @property
    def total_paid(self):
        return self.subtotal_paid + self.shipping_paid
